(* +------------------------------------------------------------------------+
   |                                                                        |
   |                    Edition des images et du cache                      |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

open Global
open Graphics
open Widgets
open Calc

                 (* +-----------------------------+
                    |  Conversion de coordonnes  |
                    +-----------------------------+ *)

let ints_of_floats f img (x,y) =
  truncate((x -. img.x0)*.float(img.sx)/.img.dx) + f.fx + f.fw/2,
  truncate((y -. img.y0)*.float(img.sy)/.img.dy) + f.fy + f.fh/2
  
let floats_of_ints f img (x,y) =
  img.x0 +. img.dx*.float(x - f.fx - f.fw/2)/.float(f.fw),
  img.y0 +. img.dy*.float(y - f.fy - f.fh/2)/.float(f.fh)


                    (* +-----------------------+
                       |  dition d'une image  |
                       +-----------------------+ *)


let edit_image master data nature =

  let {fx=x; fy=y; fw=w; fh=h} = data.fdraw and d = fl_width + 4 in
  let boutons = 
    (make_fleches {fx=x-d; fy=y-d; fw=w+2*d; fh=h+2*d})
    @ (make_boutons data.fmsg
	 (["agrandir"; "rtrcir"; "recentrer"]
	  @ (match nature with MANDELBROT -> ["julia"] | _ -> [])
	  @ ["enregistrer"; "quitter"]))
  in
  List.iter (draw_bouton false) boutons;

  let get_img() = match nature with
    | MANDELBROT -> data.mandel
    | JULIA      -> data.julia
  and set_img img = match nature with
    | MANDELBROT -> data.mandel <- img
    | JULIA      -> data.julia  <- img
  in

  let draw() =
    let img = get_img() in
    calc_image master data img;
    if img.nat = MANDELBROT then begin
      let f = data.fdraw in
      let (x,y) = ints_of_floats f img (data.julia.jx,data.julia.jy) in
      set_color sombre;
      if x >= f.fx & x < f.fx+f.fw then draw_poly [| x,f.fy; x,f.fy+f.fh-1 |];
      if y >= f.fy & y < f.fy+f.fh then	draw_poly [| f.fx,y; f.fx+f.fw-1,y |];
      set_color clair;
      let x = x+1 and y = y-1 in
      if x >= f.fx & x < f.fx+f.fw then	draw_poly [| x,f.fy; x,f.fy+f.fh-1 |];
      if y >= f.fy & y < f.fy+f.fh then	draw_poly [| f.fx,y; f.fx+f.fw-1,y |];
    end
  in

  draw();
  let res = ref(None) in
  while !res = None do
    let r = vnement false {boutons=boutons; entres=[]; onglets=data.ongl} in
    let img = get_img() in
    match r with

      | Some(BOUTON "quitter") -> res := r;
      | Some(ONGLET o) -> if o <> (List.hd data.ongl).titre then res := r

      | Some(FLECHE(dir)) ->
	  let (x0,y0) = match dir with
	    |   0  -> img.x0 +. 0.5*.img.dx, img.y0
	    |   45 -> img.x0 +. 0.5*.img.dx, img.y0 +. 0.5*.img.dy
	    |   90 -> img.x0,                img.y0 +. 0.5*.img.dy
	    |  135 -> img.x0 -. 0.5*.img.dx, img.y0 +. 0.5*.img.dy
	    |  180 -> img.x0 -. 0.5*.img.dx, img.y0
	    |  -45 -> img.x0 +. 0.5*.img.dx, img.y0 -. 0.5*.img.dy
	    |  -90 -> img.x0,                img.y0 -. 0.5*.img.dy
	    | -135 -> img.x0 -. 0.5*.img.dx, img.y0 -. 0.5*.img.dy
	    | _    -> failwith "cas impossible"
	  in
	  set_img {img with x0=x0; y0=y0};
	  draw()

      | Some(BOUTON "agrandir") -> begin
	  match select_rectangle data.fmsg data.fdraw with
	    | None -> draw()
	    | Some(x,y,r) ->
		set_img {img with
		  x0 = img.x0 +. x*.img.dx;
		  y0 = img.y0 +. y*.img.dy;
		  dx = img.dx*.r;
		  dy = img.dy*.r};
		draw()
	end

      | Some(BOUTON "rtrcir") -> begin
	  match select_rectangle data.fmsg data.fdraw with
	    | None -> draw()
	    | Some(x,y,r) ->
		set_img {img with
		  x0 = img.x0 -. x*.img.dx/.r;
		  y0 = img.y0 -. y*.img.dy/.r;
		  dx = img.dx/.r;
		  dy = img.dy/.r};
		draw()
	end

      | Some(BOUTON "recentrer") -> begin
	  List.iter (fun b -> if b.bevt = BOUTON "recentrer"
		     then draw_bouton true b) boutons;
	  match vnement false {boutons=[]; onglets=[]; entres=[]} with
	    | Some(POINT(x,y)) when inside data.fdraw x y ->
		let (x,y) = floats_of_ints data.fdraw img (x,y) in
		set_img {img with x0=x; y0=y};
		draw()
	    | _ -> List.iter (draw_bouton false) boutons
	end

      | Some(BOUTON "julia") -> begin
	  List.iter (fun b -> if b.bevt = BOUTON "julia"
		     then draw_bouton true b) boutons;
	  match vnement false {boutons=[]; onglets=[]; entres=[]} with
	    | Some(POINT(x,y)) when inside data.fdraw x y ->
		let (x,y) = floats_of_ints data.fdraw img (x,y) in
		data.julia <- {data.julia with jx=x; jy=y};
		res := Some(ONGLET "Julia")
	    | _ -> List.iter (draw_bouton false) boutons
	end

      | Some(BOUTON "enregistrer") -> begin
	  try
	    let pixmap = List.assoc img data.mem in
	    File.enregistre data pixmap
	  with Not_found -> ()
	end

      | Some(RESIZE) -> res := r;
      | _ -> ()

  done;
  !res

                  (* +--------------------------+
                     |  Visualisation du cache  |
                     +--------------------------+ *)

let visu_cache data =

  let {fx=x; fy=y; fw=w; fh=h} = data.fdraw in
  draw_contour true {fx=x-4; fy=y-4; fw=w+8; fh=h+8};

  let boutons =
    make_boutons data.fmsg
      ["prcdent"; "suivant"; "valider"; "enregistrer"; "quitter"] in
  List.iter (draw_bouton false) boutons;

  let cache = Array.of_list data.mem in
  let n = Array.length(cache)
  and i = ref(0) in

  let res = ref(None) in
  while !res = None do
    if !i < n then draw_image (snd(cache.(!i))) data.fdraw.fx data.fdraw.fy;
    let r = vnement false {boutons=boutons; entres=[]; onglets=data.ongl} in
    match r with
      | Some(BOUTON "prcdent") -> if !i < n-1 then incr i
      | Some(BOUTON "suivant")   -> if !i > 0   then decr i
      | Some(BOUTON "valider")   -> if !i < n then begin
	  let img = fst cache.(!i) in
	  match img.nat with
	    | MANDELBROT -> data.mandel <- img; res := Some(ONGLET "Mandelbrot")
	    | JULIA      -> data.julia  <- img; res := Some(ONGLET "Julia")
	end
      | Some(BOUTON "enregistrer") when !i < n ->
	  File.enregistre data (snd cache.(!i))
      | Some(BOUTON "quitter")   -> res := r
      | Some(ONGLET o) -> if o <> (List.hd data.ongl).titre then res := r
      | Some(RESIZE) -> res := r;
      | _ -> ()
  done;
  !res



