(* +------------------------------------------------------------------------+
   |                                                                        |
   |                       Elments de l'interface graphique                |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

open Global
open Graphics

(* trac de ligne polygonale, oubli dans graphics *)
let draw_poly v =
  let (x,y) = v.(0) in moveto x y;
  for i=1 to Array.length(v)-1 do let (x,y) = v.(i) in lineto x y done


                         (* +------------+
                            |  Fentres  |
                            +------------+ *)

(* trace un contour en relief *)
let draw_contour creux {fx=x;fy=y;fw=w;fh=h} =

  set_color (if creux then clair else sombre);
  draw_poly [| x,(y+2); (x+2),y; (x+w-2),y; (x+w-1),(y+2); (x+w-1),(y+h-3); (x+w-3),(y+h-1)|];
  set_color (if creux then sombre else clair);
  draw_poly [| x,(y+2); x,(y+h-3); (x+2),(y+h-1); (x+w-3),(y+h-1) |]

(* remplit une fentre avec la couleur fournie *)
let fill_fentre {fx=x;fy=y;fw=w;fh=h} couleur =

  set_color couleur;
  let p = [|
    x,       (y+2);
    (x+2),   y;
    (x+w-2), y;
    (x+w),   (y+2);
    (x+w),   (y+h-2);
    (x+w-2), (y+h);
    (x+2),   (y+h);
    x,       (y+h-2);
    x,       (y+2)
  |]
  in
  fill_poly p; draw_poly p

(* dit si un point est dans une fentre *)
let inside {fx=x;fy=y;fw=w;fh=h} x0 y0 =
  x0 >= x & y0 >= y & x0 < x+w & y0 < y+h 

(* copie d'cran *)
let get_image {fx=x;fy=y;fw=w;fh=h} = Graphics.get_image x y w h

                          (* +-----------+
                             |  Boutons  |
                             +-----------+ *)

(* dessine un bouton *)
let draw_bouton creux bouton =

  fill_fentre bouton.bf fond_actif;
  draw_contour creux bouton.bf;
  set_color crayon;
  bouton.draw bouton.bf

(* cre un bouton texte *)
let bouton_texte(fentre,texte) =
  let l = ref(fst(text_size texte))
  and i = ref(String.length texte) in
  while !i > 0 & !l > fentre.fw-5 do
    decr i;
    l := fst(text_size (String.sub texte 0 !i))
  done;
  let t = String.sub texte 0 !i in
  let draw {fx=x; fy=y; fw=w; fh=h} =
    moveto (x + (w - !l)/2) (y+2);
    draw_string t
  in {bf=fentre; clic=false; bevt=BOUTON(texte); draw=draw}

(* cre une range de boutons texte *)
let make_boutons fentre boutons =

  let l = List.length(boutons)
  and fw = fentre.fw+1
  in
  let rec mb n = function
    | [] -> []
    | h::t ->
	let x = fw*n/l         in
	let w = fw*(n+1)/l - x in
	(bouton_texte({fentre with fx=fentre.fx+x; fw=w-1},h)) :: (mb (n+1) t)
  in mb 0 boutons

(* cre un bouton avec une flche *)
let bouton_fleche(fentre,orient) =

  let a = 4.0*.atan(1.0)*.float(orient)/.180.0 in
  let c = cos(a)
  and s = sin(a)
  and f = 0.6*.float(fl_width)in
  let tr(x,y) = truncate(f*.(c*.x -. s*.y)),
                truncate(f*.(s*.x +. c*.y))  in
  let p = Array.map tr [| 0.577,0.0; -0.289,0.5; -0.289,-0.5; 0.577,0.0 |] in
  let draw {fx=x; fy=y; fw=w; fh=h} =
    let x0 = x+w/2 and y0 = y+h/2 in
    fill_poly (Array.map (fun (u,v) -> u+x0,v+y0) p)
  in {bf=fentre; clic=false; bevt=FLECHE(orient); draw=draw}

(* cre des flches dans toutes les directions *)
let make_fleches(fentre) =

  let x0 = fentre.fx              in
  let x1 = x0 + fl_width + 1       in
  let x3 = fentre.fx + fentre.fw in
  let x2 = x3 - fl_width           in

  let y0 = fentre.fy              in
  let y1 = y0 + fl_width + 1       in
  let y3 = fentre.fy + fentre.fh in
  let y2 = y3 - fl_width           in

  List.map bouton_fleche [
    {fx=x0; fy=y0; fw=fl_width; fh=fl_width}, -135;
    {fx=x0; fy=y1; fw=fl_width; fh=y2-y1-1 },  180;
    {fx=x0; fy=y2; fw=fl_width; fh=fl_width},  135;
    {fx=x1; fy=y0; fw=x2-x1-1;  fh=fl_width},  -90;
    {fx=x1; fy=y2; fw=x2-x1-1;  fh=fl_width},   90;
    {fx=x2; fy=y0; fw=fl_width; fh=fl_width},  -45;
    {fx=x2; fy=y1; fw=fl_width; fh=y2-y1-1 },    0;
    {fx=x2; fy=y2; fw=fl_width; fh=fl_width},   45 ]


(* enfonce ou relche un bouton *)
let actionne_bouton res x y down bouton =
  let dedans = inside bouton.bf x y in
  let pouss = dedans & down        in
  if bouton.clic <> pouss then begin
    bouton.clic <- pouss;
    draw_bouton pouss bouton;
    if dedans & not(down) then res := Some(bouton.bevt)
  end

                     (* +--------------------+
                        |  Champs de saisie  |
                        +--------------------+ *)

(* cre une colonne de champs *)
let rec make_entres w h x0 y0 = function
  | [] -> []
  | l::suite ->
      {ef={fx=x0; fy=y0; fw=w; fh=h}; leg=l; texte=""}
      :: (make_entres w h x0 (y0-h) suite)
  
(* affiche un champ *)
let draw_entre curseur {ef=ef; leg=leg; texte=texte} =

  fill_fentre ef fond_actif;
  set_color crayon;
  moveto (ef.fx + 2) (ef.fy + 2);
  draw_string leg;
  if curseur >= 0 then begin
    draw_string (String.sub texte 0 curseur);
    let (x,y) = current_point() in
    fill_poly [| x,(y+4); (x-3),(y-2); (x+3),(y-2); x,(y+4) |];
    draw_string (String.sub texte curseur
		   (String.length(texte) - curseur));
    draw_contour true ef
  end
  else draw_string texte

(* traite une touche et retourne la nouvelle position du curseur *)
let handle_key c k e =
  let t = e.texte in
  let l = String.length(t)
  and lmax = en_width - String.length(e.leg) in
  match k with
  | (' '..'\255') when l < lmax ->
      e.texte <- (String.sub t 0 c) ^ (String.make 1 k) ^ (String.sub t c (l-c));
      c+1
  | '\004' when c < l ->
      e.texte <- (String.sub t 0 c) ^ (String.sub t (c+1) (l-c-1));
      c
  | '\008' when c > 0 ->
      e.texte <- (String.sub t 0 (c-1)) ^ (String.sub t c (l-c));
      c-1
  | '\011' ->
      e.texte <- String.sub t 0 c;
      c
  | '\021' ->
      e.texte <- String.sub t c (l-c);
      0
  | '\012' ->
      e.texte <- "";
      0
  | '\001' -> 0
  | '\005' -> l
  | '\002' when c > 0 -> c-1
  | '\006' when c < l -> c+1
  | _ -> c

(* cherche quel caractre est point *)      
let locate_cursor x e =

  let t = e.leg ^ e.texte      in
  let i = String.length(e.leg) in
  let j = String.length(t)     in
  let x = x - e.ef.fx - 2      in
  let c = ref(x/fst(text_size "M")) in
  if !c < i then c := i else if !c > j then c := j;
  while !c > i & x < fst(text_size(String.sub t 0 !c)) - 2 do decr c done;
  while !c < j & x > fst(text_size(String.sub t 0 !c)) + 2 do incr c done;
  !c - i


(* texte dit sans les espaces de tte ni de fin *)
exception ERREUR of entre * string
let strip(entre) =
  let i = ref(0) and j = ref(String.length entre.texte) in
  while !i < !j & entre.texte.[!i] = ' '   do incr i done;
  while !j > !i & entre.texte.[!j-1] = ' ' do decr j done;
  if !i < !j then String.sub entre.texte !i (!j - !i)
  else raise (ERREUR(entre,"champ vide"))



                               (* +-----------+
                                  |  Onglets  |
                                  +-----------+ *)


(* cre une range d'onglets *)
let make_onglets fentre titres =
  let y = fentre.fy + fentre.fh - oheight in
  let rec mo x = function
    | [] -> []
    | t::suite ->
	{ow={fx=x; fy=y; fw=owidth; fh=oheight}; titre=t} :: (mo (x+owidth) suite)
  in mo (fentre.fx + pad + 8) titres

(* affiche un onglet en premier plan ou en arrire plan *)
let draw_onglet devant onglet =

  let x3 = onglet.ow.fx in
  let x2 = x3 - 2       in
  let x1 = x2 - 4       in
  let x0 = x1 - 2       in
  let x4 = x0 + onglet.ow.fw in
  let x5 = x4 + 2       in
  let x6 = x5 + 4       in
  let x7 = x6 + 2       in

  let y0 = onglet.ow.fy -1   in
  let y1 = y0 + 2            in
  let y3 = y0 + onglet.ow.fh in
  let y2 = y3 - 2            in

  set_color (if devant then fond_actif else fond_passif);
  fill_poly [|x0,y0; x1,y1; x2,y2; x3,y3; x4,y3; x5,y2; x6,y1; x7,y0; x0,y0|];
  if devant then draw_poly [| x0,y0; x7,y0 |];
  set_color sombre;
  draw_poly [|x7,y0; x6,y1; x5,y2; x4,y3|];
  set_color clair;
  draw_poly [|x4,y3; x3,y3; x2,y2; x1,y1; x0,y0|];

  set_color crayon;
  let w,_ = text_size(onglet.titre) in
  moveto ((x0+x7-w)/2) y1;
  draw_string onglet.titre


(* affiche les onglets et le cadre dessous *)
let draw_onglets fentre onglets =
  let f = {fentre with fh = fentre.fh - oheight} in
  List.iter (draw_onglet false) (List.rev(List.tl onglets));
  fill_fentre f fond_actif;
  draw_contour false f;
  draw_onglet true (List.hd onglets)



                    (* +-----------------------+
                       |  Attend un vnement  |
                       +-----------------------+ *)

let vnement poll widgets =
  if not(poll) & ((size_x() <> !wx) or (size_y() <> !wy))
  then Some(RESIZE)
  else begin
    let res  = ref(None) in
    let evts = if poll
    then [Poll] else [Button_down; Button_up; Key_pressed] in
    let {mouse_x=x; mouse_y=y; button=b; keypressed=k; key=k'} =
      wait_next_event(evts) in
    List.iter (actionne_bouton res x y b) widgets.boutons;
    if !res = None & b then
      List.iter
	(fun o -> if inside o.ow x y then res := Some(ONGLET o.titre))
	widgets.onglets;
    if !res = None & b then res := Some(POINT(x,y));
    if !res = None & k then res := Some(TOUCHE(k'));
    !res
  end


                          (* +---------------------+
                             |  dition de champs  |
                             +---------------------+ *)

let dite widgets e0 =

  let e = Array.of_list widgets.entres in
  let n = Array.length e in
  let i = ref(n-1)
  and c = ref(0)
  and res = ref(None) in

  while !i > 0 & e.(!i) <> e0 do decr i done;
  for j=0 to n-1 do if j <> !i then draw_entre (-1) e.(j) done;
  draw_entre 0 e.(!i);
  List.iter (draw_bouton false) widgets.boutons;

  while !res = None do
    match vnement false widgets with
      | Some(BOUTON(_)) as b -> res := b
      | Some(ONGLET(_)) as b -> res := b
      | Some(RESIZE)    as b -> res := b
      | Some(POINT(x,y)) ->
	  let j = ref(0) in
	  while !j < n & not(inside e.(!j).ef x y) do incr j done;
	  if !j < n then begin
	    if !j <> !i then begin
	      draw_entre (-1) e.(!i);
	      i := !j
	    end;
	    c := locate_cursor x e.(!i);
	    draw_entre !c e.(!i)
	  end
      | (Some(TOUCHE('\013')) | Some(TOUCHE('\009')) | Some(TOUCHE('\014'))) ->
	  draw_entre (-1) e.(!i);
	  i := (!i + 1) mod n;
	  c := 0;
	  draw_entre 0 e.(!i)
      | Some(TOUCHE('\016')) ->
	  draw_entre (-1) e.(!i);
	  i := (!i + n - 1) mod n;
	  c := 0;
	  draw_entre 0 e.(!i)
      | Some(TOUCHE(k)) ->
	  c := handle_key !c k e.(!i);
	  draw_entre !c e.(!i)
      | _ -> ()
  done;
  !res

(* attend une entre non vide valide par Entre ou Esc *)
let ask entre =

  let widgets = {entres=[entre]; onglets=[]; boutons=[]} in
  let c       = ref(0)
  and encore  = ref(true)
  and res     = ref(None)
  in

  let pixmap = get_image entre.ef in
  draw_entre 0 entre;
  while !encore do
    match vnement false widgets with
      | Some(POINT(x,y)) when inside entre.ef x y ->
	  c := locate_cursor x entre;
	  draw_entre !c entre
      | Some(TOUCHE('\027')) -> encore := false
      | Some(RESIZE)         -> encore := false
      | Some(TOUCHE('\013')) -> begin
	  try res := Some(strip(entre)); encore := false
	  with ERREUR(_) -> ()
	end
      | Some(TOUCHE(k)) ->
	  c := handle_key !c k entre;
	  draw_entre !c entre
      | _ -> ()
  done;
  draw_image pixmap entre.ef.fx entre.ef.fy;
  !res

                (* +----------------------------------------+
                   |  Slectionne un rectangle  la souris  |
                   +----------------------------------------+ *)

let select_rectangle fmsg f =

  let pixmap = get_image f
  and pmsg   = get_image fmsg in
  let boutons = make_boutons fmsg ["valider"; "annuler"] in
  List.iter (draw_bouton false) boutons;

  let x0 = ref(f.fx +  f.fw/2)
  and y0 = ref(f.fy +  f.fh/2)
  and dx = ref(f.fw/20)
  and dy = ref(f.fh/20) in

  let draw_rect() =
    draw_image pixmap f.fx f.fy;
    set_color clair;
    let a = !x0 - !dx + 1
    and b = !x0 + !dx
    and c = !y0 - !dy
    and d = !y0 + !dy - 1 in
    draw_poly [| a,c; a,d; b,d; b,c; a,c |];
    set_color sombre;
    let a=a-1 and b=b-1 and c=c+1 and d=d+1 in
    draw_poly [| a,c; a,d; b,d; b,c; a,c |]
  in

  let change_rect x1 y1 w h =
    if x1 <> !x0 or y1 <> !y0 or w <> !dx or h <> !dy
    then begin
      x0 := x1;
      y0 := y1;
      dx := w;
      dy := h;
      draw_rect()
    end
  in

  let drag_rect x y =
    let u0 = !x0 and v0 = !y0 in
    let _ = wait_next_event [Button_up; Mouse_motion] in
    while button_down() do
      let (u,v) = mouse_pos() in
      let x1 = min (max (u0+u-x) (f.fx + !dx)) (f.fx + f.fw - !dx - 1)
      and y1 = min (max (v0+v-y) (f.fy + !dy)) (f.fy + f.fh - !dy - 1)
      in change_rect x1 y1 !dx !dy
    done
  in

  let drag_bord () =
    while button_down() do
      let (u,v) = mouse_pos() in
      let w = max 5 (abs(u - !x0))
      and h = max 5 (abs(v - !y0))
      and a = min (!x0-f.fx) (f.fx+f.fw-1 - !x0)
      and b = min (!y0-f.fy) (f.fy+f.fh-1 - !y0) in
      let (w,h) = if w*f.fh > h*f.fw then (w,w*f.fh/f.fw) else (h*f.fw/f.fh,h)
      and (a,b) = if a*f.fh < b*f.fw then (a,a*f.fh/f.fw) else (b*f.fw/f.fh,b)
      in change_rect !x0 !y0 (min w a) (min h b)
    done
  in

  draw_rect();
  let encore = ref(true) and res = ref(None) in

  while !encore do
    match vnement false {boutons=boutons; entres=[]; onglets=[]} with
      | Some(POINT(x,y)) when
	    x > !x0 - !dx + 2 & x < !x0 + !dx - 2
	  & y > !y0 - !dy + 2 & y < !y0 + !dy - 2 -> drag_rect x y
      | Some(POINT(x,y)) when
	    x > !x0 - !dx - 5 & x < !x0 + !dx + 5
	  & y > !y0 - !dy - 5 & y < !y0 + !dy + 5 -> drag_bord()
      | Some(BOUTON "valider") ->
	  let x = float(!x0 - f.fx - f.fw/2)/.float(f.fw)
	  and y = float(!y0 - f.fy - f.fh/2)/.float(f.fh)
	  and r = float(!dx*2)/.float(f.fw) in
	  res := Some(x,y,r);
	  encore := false;
      | Some(BOUTON "annuler") -> encore := false;
      | _ -> ()
  done;

  draw_image pmsg fmsg.fx fmsg.fy;
  !res

