(* +------------------------------------------------------------------------+
   |                                                                        |
   |                     Paralllisation d'un calcul                        |
   |                                                                        |
   +------------------------------------------------------------------------+ *)

(* M. Quercia, mars 1999 *)


                         (* +-------------+
                            |  Dboguage  |
                            +-------------+ *)

val log   : string -> unit               (* affiche un message       *)
val debug : bool ref                     (* trace les communications *)

(*
 * [log message] affiche le message fourni sur l'erreur standard avec
 * la date.
 *
 * Si [!debug] vaut [true], les communications entre le matre et les
 * esclaves sont affiches sur l'erreur standard.
 * Initialement, [!debug] vaut [false].
 *)

exception Job_wait
exception Job_giveup

                          (* +-----------+
                             |  Esclave  |
                             +-----------+ *)

module Slave : sig

  type ('init,'job,'result) slave

  val connect : string ->
    ('init   -> string) ->          
    ('job    -> string) ->
    ('result -> string) ->
    ('init -> ('init,'job,'result) slave -> ('job -> 'result)) ->
    unit

  val listen : int ->
    ('init   -> string) ->          
    ('job    -> string) ->
    ('result -> string) ->
    ('init -> ('init,'job,'result) slave -> ('job -> 'result)) ->
    unit

  val inetd :
    ('init   -> string) ->          
    ('job    -> string) ->
    ('result -> string) ->
    ('init -> ('init,'job,'result) slave -> ('job -> 'result)) ->
    unit

(*
 * [connect maitre soi soj sor make_calc] contacte la machine matre
 * (chaine "host:port") puis lance les calculs. [soi,soj,sor] sont
 * les fonctions de conversion en chane des donnes reues ou transmises
 * au matre, [make_calc] est une fonction appele avec le bloc
 * d'initialisation reu du matre et le descripteur de donnes construit
 * lors de la connexion. Elle doit retourner une fonction [calc] qui sera
 * appele pour chaque travail soumis par le matre.
 *
 * [listen port soi soj sor make_calc] ouvre le port spcifi en coute
 * et attend une connection d'un matre. Pour chaque connexion entrante,
 * un processus esclave est lanc qui effectue les calculs soumis par le
 * matre. Les paramtres [soi soj sor make_calc] sont interprts de
 * la mme manire que [connect]. Cette fonction ne retourne normalement pas.
 *
 * [inetd soi soj sor make_calc] excute un unique processus esclave avec
 * une connexion dj tablie et relie aux cannaux d'entre et de sortie
 * standard.
 *)


  val try_giveup : ('init,'job,'result) slave -> unit

 (*
  * [try_giveup sdata] lve l'exception [Job_giveup] si le matre a demand
  * l'interruption du calcul en cours. La fonction de calcul est suppose
  * appeler [try_giveup] de temps en temps pour dterminer si elle doit
  * continuer son travail. En cas d'abandon, [calc] peut retourner un
  * rsultat qui sera ignor ou propager l'exception [Job_giveup] qui sera
  * rattrape. Toute autre exception est fatale.
  *)

end

                          (* +----------+
                             |  Matre  |
                             +----------+ *)

module Master: sig


  type ('init,'job,'result) master

  val init :                                       (* dmarre une session *)
    'init ->                                       (* init. esclave       *)
    (unit -> 'job) ->                              (* fournit un travail  *)
    ('job -> 'result -> 'job list -> 'job list) -> (* traite un rsultat  *)
    ('init   -> string) ->                       (* fonction d'impression *)
    ('job    -> string) ->
    ('result -> string) ->
    ('init,'job,'result) master

 (*
  * [init idata job result soi soj sor] construit un bloc de donnes associ
  *  un calcul parallle. [idata] est un bloc d'initialisation  transmettre
  *  chaque esclave; [job] est une fonction qui sera appele  chaque fois
  * qu'il faudra donner du travail  un esclave; [result] est une fonction
  * qui sera appele  chaque fois qu'une machine esclave retournera un
  * rsultat; [soi] [soj] et [sor] sont des fonctions de conversion des
  * donnes en chanes utilises en cas de dboguage.
  *)

  val listen  : ('a,'b,'c) master -> int    -> unit
  val connect : ('a,'b,'c) master -> string -> unit
  val autonomous : ('init,'job,'result) master -> ('job -> 'result) -> unit

 (*
  * Association de machines esclaves au calcul parallle.
  *
  * L'enregistrement par la machine matre d'une machine esclave pour
  * participation  un calcul parallle est effectu lorsque l'une des deux
  * machines contacte l'autre.
  *
  * Le matre attend les connexions des esclaves :
  * ----------------------------------------------
  * [listen mdata port] ouvre sur la machine matre le port TCP [port] en
  * coute, les machines esclaves peuvent se connecter sur ce port de leur
  * propre initiative pour se faire enregistrer.
  *
  * Le matre contacte les esclaves :
  * ---------------------------------
  * [connect mdata esclave] : le matre contacte la machine esclave
  * spcifi (chane "host:port").
  *
  * Dans les deux cas, une fois la connexion tablie, le matre transmet
  *  l'esclave le bloc d'initialisation puis les travaux  effectuer et
  * rceptionne les rsultats. Les deux mthodes d'association peuvent
  * tre utilises concurrement (le matre contacte certains esclaves
  * prdfinis et reoit les connexions d'autres esclaves ventuels), et
  * il est aussi possible de dfinir plusieurs ports d'coute par des
  * appels successifs  [listen].
  *
  * Le matre simule un esclave :
  * -----------------------------
  * [autonomous mdata calc] : le matre dsactive le mode parallle et
  * excute lui-mme tous les calculs retourns par [job].
  *
  * Des appels successifs  [listen], [connect] et [autonomous] basculent
  * le matre entre les modes parallle et autonome. Lors du passage en
  * mode autonome, toutes les connexions ouvertes en mode parallle sont
  * refermes.
  *)

  val start : ('a,'b,'c) master -> unit
  val stop  : ('a,'b,'c) master -> unit

 (*
  * [start mdata] dmarre une nouvelle srie de calculs. Pour chaque machine
  * esclave inoccupe, le matre prend un nouveau travail par appel  la
  * fonction [job] et le transmet  la machine esclave. Si [job] lve
  * l'exception [Job_wait], le matre soumet  l'esclave un travail en cours
  * pour lequel aucun rsultat n'a encore t reu s'il y en a.
  * Lorsqu'une machine esclave retourne un rsultat, le matre appelle la
  * fonction [result] avec le travail, le rsultat et la liste des autres
  * travaux en cours. [result] doit retourner la liste des travaux 
  * interrompre. [start] retourne lorsque tous les travaux sont termins
  * ([job] a lev l'exception [Job_wait] et la liste des travaux en cours
  * est vide).
  *
  * [stop mdata] termine la session parallle en transmettant  chaque machine
  * esclave une commande de terminaison et en fermant les ports TCP ouverts
  * par [listen]. Le bloc de donnes [mdata] reste utilisable pour une
  * nouvelle session parallle relanant les fonctions [listen] ou [connect].
  *)

end

