(* +------------------------------------------------------------------------+
   |                                                                        |
   |        Recherche parallle du premier nombre premier  N chifres       |
   |                                                                        |
   | programme principal                                                    |
   +------------------------------------------------------------------------+ *)

open Printf
open Parallele
open Rabin

                               (* +----------+
                                  |  Matre  |
                                  +----------+ *)

let maitre () =

  let port   = ref(0)
  and n      = ref(0)
  and k      = ref(0)
  and b      = ref(0)
  and croit  = ref true
  and bases  = ref []
  and slaves = ref []
  and help   = ref false
  and auto   = ref false  in

  let (=::) l x = l := x :: !l in

  let argspec = [
    "-a", Arg.Int((=::) bases), "<int> base pour Rabin-Miller";
    "-b", Arg.Int((:=) b),      "<int> limine les diviseurs premiers <= b";
    "-c", Arg.Set(croit),      	"recherche croissante";
    "-d", Arg.Clear(croit),     "recherche dcroissante";
    "-h", Arg.Set(help),        "liste des options";
    "-k", Arg.Int((:=) k),      "<int> commence  10^n+k";
    "-n", Arg.Int((:=) n),      "<int> commence  10^n+k";
    "-p", Arg.Int((:=) port),   "<int> port TCP";
    "-s", Arg.String((=::) slaves), "<string:int> esclave  contacter";
    "-auto", Arg.Set(auto),     "mode autonome";
    "-v", Arg.Set(debug),      	"trace les communications"]

  and anon(s) = raise (Arg.Bad ("paramtre non reconnu : " ^ s))
  and msg ="syntaxe : rabin maitre <paramtres>"
  in

  Arg.current := 1;
  Arg.parse argspec anon msg;
  
  if !help then begin
    Arg.usage argspec msg;
    exit(1)
  end;

  let data = {n0      = Num.power_num (Num.Int 10) (Num.Int !n);
              p       = pprime(!b);
	      op      = if !croit then Num.add_num else Num.sub_num;
              bases   = !bases;
	      knext   = if !k mod 2 = 1 then !k else !k+1;
	      a_faire = [];
	      kfound  = -1} in

  let m = Master.init (!n,!croit) (job data) (result data) soi soj sor in

  if !port <> 0 then Master.listen m !port;
  List.iter
    (fun s -> try Master.connect m s
              with _ -> log (sprintf "unable to connect to %s\n" s))
    !slaves;
  if !auto then Master.autonomous m (rabin data.op (fun () -> ()) data.n0);

  Master.start m;
  Master.stop m;

  printf "%d\n" data.kfound;
  flush stdout;
  exit(0)


                               (* +-----------+
                                  |  Esclave  |
                                  +-----------+ *)

let esclave() =

  let maitre = ref ""
  and port   = ref(0)
  and help   = ref false in

  let argspec = [
    "-h", Arg.Set(help),           "liste des options";
    "-m", Arg.String((:=) maitre), "<string:int> matre  contacter";
    "-p", Arg.Int((:=) port),      "<int> port TCP";
    "-v", Arg.Set(debug),          "trace les communications"
  ]
  and anon(s) = raise (Arg.Bad ("paramtre non reconnu : " ^ s))
  and msg     = "syntaxe : rabin esclave <paramtres>" in

  Arg.current := 1;
  Arg.parse argspec anon msg;
  
  if !help then begin
    Arg.usage argspec msg;
    exit(1)
  end;

  let make_calc (n,c) s =
    let n0 = Num.power_num (Num.Int 10) (Num.Int n)
    and op = if c then Num.add_num else Num.sub_num
    in rabin op (fun () -> Slave.try_giveup s) n0
  in
    
  if !maitre <> ""   then Slave.connect !maitre soi soj sor make_calc
  else if !port <> 0 then Slave.listen  !port   soi soj sor make_calc
                     else Slave.inetd           soi soj sor make_calc;
  exit(0)
;;

                         (* +-----------------------+
                            |  Programme principal  |
                            +-----------------------+ *)


begin

  match Array.to_list Sys.argv with
    | _ :: "maitre"   :: _ -> maitre()
    | _ :: "esclave"  :: _ -> esclave()
    | _ ->
	fprintf stderr "syntaxe : rabin maitre   <paramtres>\n";
	fprintf stderr "          rabin esclave  <paramtres>\n";
	fprintf stderr "rabin (maitre|esclave) -h dtaille les paramtres\n";
	flush stderr;
	exit(1)

end
